/*
 * sqlget.c Retrieve a database item.
 *
 * Based on the example given in:
 * /usr/share/doc/postgresql-7.2.1/html/libpq-example.html
 *
 * Richard Stover
 * February, 2003
 *
 */

#define RCSID static char *rcsid __attribute__ ((unused))
RCSID = "$Id: sqlget.c,v 1.1.1.1 2007/12/19 18:22:47 cvs Exp $";
#if 0
    $Log: sqlget.c,v $
    Revision 1.1.1.1  2007/12/19 18:22:47  cvs
    Initial checkin

    Revision 1.5  2006/05/31 20:11:02  sla
    sync with CCDLAB

    Revision 1.5  2006/03/25 01:01:32  ccdev
    freetuple() was freeing colname even though this was not set
    when there was some type of database error. Now all errors set
    colname. Also freetuple checks to see if colname is non-NULL
    before freeing it.   rjs

    Revision 1.4  2005/07/28 18:58:52  ccdev
    Was doing a BEGIN without a COMMIT when an error occurred. Now
    a COMMIT is always done if a BEGIN is done.

    Revision 1.3  2005/03/03 22:44:51  ccdev
    Prototype definition for flogerror().

    Revision 1.2  2005/03/03 22:42:50  ccdev
    Handle several different kinds of SELECTs. Use thread semaphores
    to make the database accesses thread-safe. Various other
    enhanncements.

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libpq-fe.h>
#include <semaphore.h>
#include "sqlget.h"

struct pgconnections {
    char *pghost;              /* host name of the backend server */
    char *pgport;              /* port of the backend server */
    char *pgoptions;           /* special options to start up the backend
                                 * server */
    char *pgtty;               /* debugging tty for the backend server */
    char *dbName;
    PGconn *conn;
};

void flogerror(char *f, ...);

#define MAXCON 20
static struct pgconnections pgcon[MAXCON];

/*	Close all connections to databases.				*/
void exit_sql(void)
{
    int i;
    static sem_t sem;
    static int sem_is_init = 0;

/*  Access to the database does not seem to be thread-safe.	*/
/*  So here we make sure that only one thread is talking to	*/
/*  the database at one time.					*/

/*  Initialize our thread semaphore if that has not been done.	*/

    if(sem_is_init == 0) {
	sem_is_init++;
    	sem_init(&sem,0,1);
    }

/*  Wait if we are not the first thread in this routine.	*/
/*  If we do block here, a sem_post(&sem) will free us later.	*/

    sem_wait(&sem);

    for(i=0; i<MAXCON; i++) {
	if(pgcon[i].dbName != NULL) {
    	    PQfinish(pgcon[i].conn);
	    pgcon[i].dbName = NULL;
	}
    }
    sem_post(&sem);
}

/*	Retrieve a database item
 *	action  = Type of query to make
 *	dbname	= The database
 *	table	= The table name
 *	row	= The identifier to select the row of the table
 *	column	= The name of column of the item to be retrieved.
 *	rowkey	= The name of the column in which the 'row' identifier is
 *		  matched.
 *	result	= The address of a char array in which the result is
 *		  returned.
 *	nchar	= The size of the char array 'result'. This should be
 *		  at least 64, to hold possible error messages.
 *
 * 	If there is an error the function returns -1 and an error
 * 	message is stored in 'result'.
 *	Otherwise the function returns 0 and the value retrieved from
 *	the database is returned in 'result'.
*/

int sqlget(char *action, char *dbname,char *table,char *row,char *column,
    char *rowkey,struct tuple *result,int ntuple)
{
    static char *query;
    int         i;
    char *rowkeyname;
    int ret_tuple,ret_ncol;
    int nt,nc;
    int nresult = 0;
    int debug = 0;
    static sem_t sem;
    static int sem_is_init = 0;
    PGconn     *conn;
    PGresult   *res;
    int ret = 0;

/*  Access to the database does not seem to be thread-safe.	*/
/*  So here we make sure that only one thread is talking to	*/
/*  the database at one time.					*/

/*  Initialize our thread semaphore if that has not been done.	*/

    if(sem_is_init == 0) {
	sem_is_init++;
    	sem_init(&sem,0,1);
    }

/*  Wait if we are not the first thread in this routine.	*/
/*  If we do block here, a sem_post(&sem) will free us later.	*/

    sem_wait(&sem);
    
/*  See if we are already connected to the requested database.			*/

    for(i=0; i<MAXCON; i++) {
	if (pgcon[i].dbName == NULL) break;
	if (strcmp(pgcon[i].dbName,dbname) == 0) break;
    }

    if (i == MAXCON) {
        result->value = malloc(64);
	sprintf(result->value,"Too many (%d) connections to database\n",MAXCON);
	result->colname = strdup("Error_Message");
	result++;
	result->value = (char *)0;
	sem_post(&sem);
	return -1;
    }

/*  Open a connection if we don't have one for this database.			*/

    if (pgcon[i].dbName == NULL) {
	pgcon[i].pghost = NULL;
        pgcon[i].pgport = NULL;      /* port of the backend server */
        pgcon[i].pgoptions = NULL;   /* special options to start up the backend
                                      * server */
        pgcon[i].pgtty = NULL;       /* debugging tty for the backend server */

/*	make a connection to the database */
        conn = PQsetdb(pgcon[i].pghost, pgcon[i].pgport, pgcon[i].pgoptions,
		pgcon[i].pgtty, dbname);

/*	check to see that the backend connection was successfully made	*/
        if (PQstatus(conn) == CONNECTION_BAD)
    	{
	    result->value = malloc(100);
            sprintf(result->value, "Connection to database '%s' failed.\n%s",
		dbname,PQerrorMessage(conn));
	    result->colname = strdup("Error_Message");
	    result++;
	    result->value = (char *)0;
	    sem_post(&sem);
	    return -1;
        }

/*	Save the database name and the connection.		*/
        pgcon[i].dbName = strdup(dbname);
	pgcon[i].conn = conn;
    }
    conn = pgcon[i].conn;
    
    query = malloc(512);
    /* start a transaction block */
    res = PQexec(conn, "BEGIN");
    if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
    {
	result->value = malloc(100);
        sprintf(result->value, "BEGIN command failed\n");
	result->colname = strdup("Error_Message");
	result++;
	result->value = (char *)0;
	ret = -1;
	goto done;
    }

    /*
     * should PQclear PGresult whenever it is no longer needed to avoid
     * memory leaks
     */
    PQclear(res);

    if((*rowkey == '_') && (strcmp(rowkey,"__STD") == 0)) {
	rowkeyname = "name0x3b2d29";
    } else {
	rowkeyname = rowkey;
    }
    if((*action == 'i') && (strcmp(action,"items") == 0)) {
        sprintf(query,"SELECT %s FROM %s",rowkeyname,table);
    } else if((*action == 'n') && (strcmp(action,"nitems") == 0)) {
        sprintf(query,"SELECT count(%s) FROM %s",rowkeyname,table);
    } else if((*action == 's') && (strcmp(action,"select") == 0)) {
        sprintf(query,"SELECT %s FROM %s where %s='%s'",
	    column,table,rowkeyname,row);
    } else if((*action == 'v') && (strcmp(action,"vselect") == 0)) {
        sprintf(query,"SELECT %s FROM %s where %s='%s'",
	    column,table,rowkeyname,row);
	debug = 1;
    }
    if(debug)
        flogerror("Doing query: %s",query);

    res = PQexec(conn, query);
    if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
    {
	result->value = malloc(strlen(PQerrorMessage(conn))+strlen(query)+20);
//	printf("Query %s failed\n%s",query,PQerrorMessage(conn));
        sprintf(result->value,
		"Query %s failed\n%s",query,PQerrorMessage(conn));
	result->colname = strdup("Error_Message");
	result++;
	result->value = (char *)0;
	ret = -1;
	goto done;
    }
    ret_tuple = PQntuples(res);
    ret_ncol = PQnfields(res);
    if(debug)
        flogerror("Database has %d responses with %d columns",
		ret_tuple,ret_ncol);

    for(nt=0; nt<ret_tuple; nt++) {
    if(debug)
        flogerror("Save result for tuple %d",nt);
	if(nresult < (ntuple-1)) {
	    for(nc=0; nc<ret_ncol; nc++) {
	        result->row = nt;
	        result->col = nc;
	        result->value = strdup(PQgetvalue(res,nt,nc));
	        result->colname = strdup(PQfname(res,nc));
    if(debug)
        flogerror("Save result %d: %s",nresult,result->value);
		result++;
		nresult++;
		if(nresult >= (ntuple-1)) break;
	    }
	}
	result->value = (char *)0;
    }

/*  It is an error to not have any results.			*/
    if(ret_tuple==0) {
        result->value = malloc(strlen(query)+40);
        sprintf(result->value,
                "Query %s returned no results\n",query);
	result->colname = strdup("Error_Message");
        result++;
        result->value = (char *)0;
        ret = -1;

    }

done:
    PQclear(res);

    /* commit the transaction */
    res = PQexec(conn, "COMMIT");
    PQclear(res);
    free(query);
    sem_post(&sem);

    return ret;

}

/*	Free the memory associated with a tuple, but not the tuple itself.	*/
void freetuple(struct tuple *tp)
{
        while(tp->value != (char *)0) {
            free(tp->value);
	    if(tp->colname != (char *)0)
                free(tp->colname);
            tp++;
        }
}

